import logging
from pathlib import Path

import fire
import numpy as np
from PIL import Image

from src import cpu, gpu

logger = logging.getLogger(__name__)


def read_img(file_path: str) -> np.ndarray:
    logger.debug(f"Reading image {file_path}")
    src_img = Image.open(file_path).convert("L")
    return np.asarray(src_img)


def write_img(img_arr: np.ndarray, img_name: str):
    Path("out_imgs").mkdir(exist_ok=True)
    res_img = Image.fromarray(img_arr).convert("L")
    filename = f"out_imgs/{img_name}.png"
    logger.debug(f"Writing image to {filename}")
    res_img.save(filename)


class CLI(object):
    def run(self, img_path: str, sigma_r: float = 75, sigma_d: float = 75):
        gpu_time = self.run_gpu(img_path, sigma_r, sigma_d)
        cpu_time = self.run_cpu(img_path, sigma_r, sigma_d)
        logger.info(f"CPU execution time: {cpu_time}")
        logger.info(f"GPU execution time: {gpu_time}")
        logger.info(f"Acceleration: {cpu_time / gpu_time:.5f}")

    def run_gpu(self, img_path: str, sigma_r: float = 75, sigma_d: float = 75):
        src_img = read_img(img_path)
        res_img, execution_time = gpu.filter_image(src_img, sigma_r, sigma_d)
        write_img(res_img, "gpu")
        return execution_time

    def run_cpu(self, img_path: str, sigma_r: float = 75, sigma_d: float = 75):
        src_img = read_img(img_path)
        res_img, execution_time = cpu.filter_image(src_img, sigma_r, sigma_d)
        write_img(res_img, "cpu")
        return execution_time


def setup_logging():
    logging.basicConfig(
        format="[%(levelname)8s] %(msg)s {%(filename)s:%(lineno)s}", level=logging.DEBUG
    )


if __name__ == "__main__":
    setup_logging()
    fire.Fire(CLI)
