#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import pkgutil
import time

import numpy as np
import pycuda.autoinit
from pycuda import driver as drv
from pycuda.compiler import SourceModule

logger = logging.getLogger(__name__)


def filter_image(image: np.ndarray, sigma_r: float, sigma_d: float):
    kernel_code = pkgutil.get_data("src", "kernel.cu").decode("UTF-8", "ignore")
    logger.debug("Read kernel code.")
    logger.debug(kernel_code)
    kernel = SourceModule(kernel_code)
    N, M = image.shape
    logger.debug("Image shape: {N}, {M}")
    result = np.zeros((N, M), dtype=np.uint32)
    block_size = (8, 8, 1)
    logger.debug(f"Using block size: {block_size}")
    grid_size = (int(np.ceil(N / block_size[0])), int(np.ceil(M / block_size[1])))
    logger.debug(f"Using block size: {block_size}")
    filt_gpu = kernel.get_function("filter")
    tex = kernel.get_texref("tex")
    tex.set_filter_mode(drv.filter_mode.LINEAR)
    tex.set_address_mode(0, drv.address_mode.MIRROR)
    tex.set_address_mode(1, drv.address_mode.MIRROR)
    drv.matrix_to_texref(image.astype(np.uint32), tex, order="C")
    start_time = time.time()
    filt_gpu(
        drv.Out(result),
        np.int32(N),
        np.int32(M),
        np.float32(sigma_d),
        np.float32(sigma_r),
        block=block_size,
        grid=grid_size,
        texrefs=[tex],
    )
    drv.Context.synchronize()
    end_time = time.time()
    execution_time = end_time - start_time
    logging.info(f"Execution time: {execution_time}")
    return result, execution_time
