#include <iostream>

#define BLOCK_SIZE 16

__global__ void gpu_mat_mul(float *a, float *b, float *out, int n) {
  int col = blockIdx.x * blockDim.x + threadIdx.x;
  int row = blockIdx.y * blockDim.y + threadIdx.y;
  if (col < n && row < n) {
    float sum = 0;
    for (int i = 0; i < n; i++) {
      sum += a[row * n + i] * b[i * n + col];
    }
    out[row * n + col] = sum;
  }
}

__host__ int main() {
  int n;      // Size of matrices
  srand(111); // Seed for random. (used for testing).
  std::cout << "Please enter matrices dims N" << std::endl << "> ";
  std::cin >> n; // User input for matrix size.
  unsigned int matrix_size = n * n;
  float *cpu_a;
  float *cpu_b;
  float *cpu_c;
  float *cpu_out; // Pointers to matrices.

  cudaMallocHost((void **)&cpu_a,
                 sizeof(float) * matrix_size); // Allocating matrix A on host.
  cudaMallocHost((void **)&cpu_b,
                 sizeof(float) * matrix_size); // Allocating matrix B on host.
  cudaMallocHost((void **)&cpu_c,
                 sizeof(float) * matrix_size); // Allocating matrix C to store
                                               // GPU multiplication result.
  cudaMallocHost((void **)&cpu_out,
                 sizeof(float) * matrix_size); // Allocating matrix C to store
                                               // CPU multiplication result.

  for (int i = 0; i < n; ++i) {
    for (int j = 0; j < n;
         ++j) { // Filling matrices with pseudo-random numbers.
      cpu_a[i * n + j] = rand() % 150;
      cpu_b[i * n + j] = rand() % 150;
    }
  }

  float cpu_time_ms, gpu_time_ms; // calculus time.

  cudaEvent_t start, stop; // Events for time measuring.

  cudaEventCreate(&start);
  cudaEventCreate(&stop);

  float *gpu_a; // GPU matrices pointers
  float *gpu_b;
  float *gpu_out;

  cudaMalloc((void **)&gpu_a,
             sizeof(float) * matrix_size); // Allocating memory on GPU.
  cudaMalloc((void **)&gpu_b, sizeof(float) * matrix_size);
  cudaMalloc((void **)&gpu_out, sizeof(float) * matrix_size);

  cudaMemcpy(gpu_a, cpu_a, sizeof(float) * matrix_size,
             cudaMemcpyHostToDevice); // Copying actual matrices to GPU.
  cudaMemcpy(gpu_b, cpu_b, sizeof(float) * matrix_size, cudaMemcpyHostToDevice);

  unsigned int grid_cols = (n + BLOCK_SIZE - 1) / BLOCK_SIZE;

  dim3 dimGrid(grid_cols, grid_cols);
  dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);

  cudaEventRecord(start, 0);
  gpu_mat_mul<<<dimGrid, dimBlock>>>(gpu_a, gpu_b, gpu_out, n);
  cudaMemcpy(cpu_c, gpu_out, sizeof(float) * matrix_size,
             cudaMemcpyDeviceToHost);
  cudaDeviceSynchronize();
  cudaEventRecord(stop, 0);

  cudaEventSynchronize(stop);

  cudaEventElapsedTime(&gpu_time_ms, start, stop);
  std::cout << "GPU evaluation time is " << gpu_time_ms << " Ms" << std::endl;

  cudaEventRecord(start, 0);
  for (int m_i = 0; m_i < n; ++m_i) {
    for (int n_i = 0; n_i < n; ++n_i) {
      for (int k_i = 0; k_i < n; ++k_i) {
        cpu_out[m_i * n + k_i] += cpu_a[m_i * n + n_i] * cpu_b[n_i * n + k_i];
      }
    }
  }

  cudaEventRecord(stop, 0);
  cudaEventSynchronize(stop);
  cudaEventElapsedTime(&cpu_time_ms, start, stop);
  std::cout << "CPU evaluation time is " << cpu_time_ms << " Ms" << std::endl;
  int mul_correct = 1;
  for (int i = 0; i < n; ++i) {
    for (int j = 0; j < n; ++j) {
      if (cpu_out[i * n + j] != cpu_c[i * n + j]) {
        std::cout << "Found incorrect values: " << cpu_out[i * n + j] << " "
                  << cpu_c[i * n + j] << std::endl;
        mul_correct = 0;
      }
    }
  }
  std::cout << "GPU acceleration: " << cpu_time_ms / gpu_time_ms << std::endl;
  if (mul_correct) {
    std::cout << "All calculations are correct" << std::endl;
  } else {
    std::cout << "Calculations are incorrect" << std::endl;
  }
}
