#include "cuda_runtime.h"
#include <thrust/copy.h>
#include <thrust/device_vector.h>
#include <thrust/generate.h>
#include <thrust/host_vector.h>
#include <thrust/random.h>

#include <cmath>
#include <iostream>

void initialize(thrust::device_vector<float> &v) {
  thrust::default_random_engine rng(123456);
  thrust::uniform_real_distribution<float> dist(0., 1.);
  for (size_t i = 0; i < v.size(); i++)
    v[i] = dist(rng);
}

float cpuPI(thrust::host_vector<float> &x, thrust::host_vector<float> &y) {
  int count = 0;
  unsigned int size = std::min(x.size(), y.size());
  for (int i = 0; i < size; i++) {
    if (sqrtf(x[i] * x[i] + y[i] * y[i]) < 1) {
      count++;
    }
  }
  std::cout << "CPU Points sum: " << count << std::endl;
  return float(count) * 4 / size;
}

struct pi_points_mapper {
  pi_points_mapper() {}

  __device__ bool operator()(const float &x, const float &y) const {
    return sqrtf(x * x + y * y) < 1;
  }
};

int main() {
  int points_count;
  float cpu_time_ms, gpu_time_ms;
  std::cout << "Please enter number of points" << std::endl << "> ";
  std::cin >> points_count;
  thrust::device_vector<float> dev_x(points_count);
  thrust::device_vector<float> dev_y(points_count);

  initialize(dev_x);
  initialize(dev_y);

  cudaEvent_t start, stop; // Events for time measuring.

  cudaEventCreate(&start);
  cudaEventCreate(&stop);

  thrust::host_vector<float> host_x = dev_x;
  thrust::host_vector<float> host_y = dev_y;

  cudaEventRecord(start, 0);
  float cpu_pi = cpuPI(host_x, host_y);
  std::cout << "PI value: " << cpu_pi << std::endl;
  cudaEventRecord(stop, 0);
  cudaEventSynchronize(stop);
  cudaEventElapsedTime(&cpu_time_ms, start, stop);
  std::cout << "CPU evaluation time is " << cpu_time_ms << " Ms" << std::endl;

  cudaEventRecord(start, 0);
  thrust::transform(dev_x.begin(), dev_x.end(), dev_y.begin(), dev_y.begin(),
                    pi_points_mapper());
  int reduced_points =
      thrust::reduce(dev_y.begin(), dev_y.end(), (int)0, thrust::plus<int>());
  cudaEventRecord(stop, 0);
  cudaEventSynchronize(stop);
  cudaEventElapsedTime(&gpu_time_ms, start, stop);
  std::cout << "GPU evaluation time is " << gpu_time_ms << " Ms" << std::endl;
  std::cout << "GPU points: " << reduced_points << std::endl;
  float gpu_pi = float(reduced_points) * 4 / points_count;
  std::cout << "GPU PI value: " << gpu_pi << std::endl;
  std::cout << "GPU acceleration: " << cpu_time_ms / gpu_time_ms << std::endl;
  return 0;
}
